var io = require('socket.io').listen(9999);
var sqlite3 = require('sqlite3').verbose();
var assert = require("assert");
var crypto = require("crypto");
var uuid = require("node-uuid");
var db = new sqlite3.Database('cryptochat.db');
var allSockets = [];
db.serialize(function() {
	db.run("CREATE TABLE if not exists direct (ip varchar(255))");
});

io.sockets.on('connection', function (socket) {
	socket.on('connection',function(data) {
		var address = socket.handshake.address;
		//check if we can connect direct to them, to our known port
		var caddress = address.address;
		//back connect to 9999
		backConnect(caddress,9999);
		//startup some diffie Hellman
		// the prime is shared by everyone                                              
		var server = crypto.getDiffieHellman('modp5');
		var prime = server.getPrime();
		server.generateKeys();
		var serverPub = server.getPublicKey('base64');
		socket.emit('data','{ "Command" : "KeyExchange", "pub" : "'+serverPub+'"}');
		socket.serverCrypto = server;
	});
	socket.on('data',function(data) {
		data = JSON.parse(data);
		if (data.Command == "KeyExchange") {
			socket.pubKey = data.pub;
			socket.secret = socket.serverCrypto.computeSecret(socket.pubKey,'base64','hex');
			socket.emit('encrypted',crypt('{"Command":"checking the microphone"}',socket.secret));
		}
	});
	socket.on('crypted',function(data) {
		data = JSON.parse(decrypt(data,socket.secret));
		console.log("Data: ");
		console.log(data);
		if (data.Command == "RequestNetworkID") {
			console.log("Getting Network ID");
			socket.uuid = uuid.v4();
			socket.emit('encrypted',crypt('{"Command" : "Uuid", "Uuid":"'+socket.uuid+'"}',socket.secret));
			allSockets[socket.uuid] = socket.id;
			console.log(allSockets);
		}
		if (data.Command == "SetUUID") {
			socket.uuid = data.Uuid;
			console.log("UUID Set now, this at some point needs a keypair " + socket.uuid);
			allSockets[socket.uuid] = socket.id;
		}
		if (data.Command == "Send") {
			var client = allSockets[data.To];
			io.sockets.socket(client).emit("encrypted",crypt('{"Command" : "Message","From":"'+data.From+'","Crypt":"'+data.Crypted+'"}',io.sockets.socket(client).secret));
		}
		if (data.Command == "RequestChat") {
			console.log("We got a request chat here");
			//shuck the request to the To:
			var client = allSockets[data.To];
			io.sockets.socket(client).emit("encrypted",crypt('{"Command" : "RequestChat","From" : "'+data.From+'","PubKey":"'+data.PubKey+'"}',io.sockets.socket(client).secret));
		}
		if (data.Command == "ShuckKey") {
			var client = allSockets[data.To];
			io.sockets.socket(client).emit("encrypted",crypt('{"Command" : "ShuckKey","From" : "'+data.From+'","PubKey":"'+data.PubKey+'"}',io.sockets.socket(client).secret));
		}
	});
	//socket.on('sendtocrypted',
});


function backConnect(address,port) {
	var ioclient = require('socket.io-client');
	var socket2 = ioclient.connect('http://'+address+':9999');
	socket2.on('error',function(err) {
		console.log("Error Connecting" + err + " " + socket2.address);
	});
	socket2.on('connect',function(data) {
		console.log("Got a conneciton");	
		//need to add to the list of direct connectables
		var host = socket2.socket.options.host;
		db.serialize(function() {
			db.run("DELETE FROM direct where ip='"+host+"'");
			db.run("INSERT INTO DIRECT (ip) values ('"+host+"')");
		});
	});
}


function decrypt(text,secret) {
        var decipher = crypto.createDecipher('aes-256-cbc',secret);
        var dec = decipher.update(text,'base64','utf8');
        dec += decipher.final('utf8');
        return dec;
}


function crypt(text,secret) {
        var assert = require("assert");
        var crypto = require("crypto");
        var cipher = crypto.createCipher('aes-256-cbc',secret);
        var crypted = cipher.update(text,'utf8','base64');
        crypted += cipher.final('base64');
        return crypted;
}
